;; Generated package description from minibuffer-header.el  -*- no-byte-compile: t -*-
(define-package "minibuffer-header" "0.2.0.20220807.80933" "Minibuffer header line" '((emacs "27.1")) :commit "87ab0ee3d9b58499a167a1bd3e181c09382630f2" :maintainer '("Nicolas P. Rougier" . "Nicolas.Rougier@inria.fr") :keywords '("convenience") :url "https://github.com/rougier/minibuffer-header")
